import re

from ..validators.account_number_validator import AccountNumberValidator
from ..validators.modulus import Modulus, Operations
from ..validators.validator import Validator


class Kid(Validator[str]):  # pylint: disable=E1136  # pylint/issues/2822
    def __init__(self, text: str) -> None:
        self.text = text
        self.score = self._calculate_score()

    @staticmethod
    def validate(input_str: str) -> bool:
        if not re.match(r"^[0-9]+$", input_str):
            return False
        check_digit = int(input_str[-1])
        kid_without_check_digit = input_str[:-1][::-1]
        modulus_10 = Modulus(10, [2, 1] * (len(kid_without_check_digit) // 2), operation=Operations.SUM)
        modulus_11 = Modulus(11, [2, 3, 4, 5, 6, 7])
        valid_length = len(kid_without_check_digit) > 1 and len(kid_without_check_digit) < 25

        return valid_length and check_digit in [
            modulus_10.calculate_check_digit(kid_without_check_digit),
            modulus_11.calculate_check_digit(kid_without_check_digit),
        ]

    def _calculate_score(self) -> int:
        length_score = len(self.text) // 2 if len(self.text) > 6 else 0
        leading_zero = 1 if self.text.startswith("0") else 0
        zero_count = -3 if self.text.count("0") / len(self.text) > 0.5 else 0
        return length_score + leading_zero + zero_count

    def __str__(self) -> str:
        return "{} : {}".format(self.text, self.score)

    def __repr__(self) -> str:
        return self.__str__()

    def is_likely(self, kid: str) -> bool:
        return (
            (self.validate(self.text) or AccountNumberValidator().validate(kid))
            and int(kid) > 0
            and 4 <= len(kid) <= 25
        )

    def is_valid(self) -> bool:
        return self.validate(self.text)
