from __future__ import division

import argparse
import os
import time

import torch
import tqdm
from torch.autograd import Variable

import wandb

from .default import checkpoint_path, data_config, model_def
from .eval import get_accracy_of_model
from .models import *
from .utils.datasets import *
from .utils.logger import *
from .utils.parse_config import *
from .utils.transforms import *
from .utils.utils import *


def load_filters(filters_list, current_dir):
    if filters_list is None:
        return None
    sepearted = list(map(lambda x: x.strip(), filters_list.split(",")))
    filters = []
    for i in sepearted:
        with open(os.path.join(current_dir, i), "r") as file:
            filters += list(
                map(lambda x: os.path.basename(x), list(filter(lambda x: len(x) > 0, (file.read()).split("\n"))))
            )
    return filters


def get_opts_raw():
    parser = argparse.ArgumentParser()
    parser.add_argument("--epochs", type=int, default=100, help="number of epochs")
    parser.add_argument("--batch_size", type=int, default=8, help="size of each image batch")
    parser.add_argument("--gradient_accumulations", type=int, default=8, help="number of gradient accums before step")
    parser.add_argument("--model_def", type=str, default=model_def, help="path to model definition file")
    parser.add_argument("--data_config", type=str, default=data_config, help="path to data config file")
    parser.add_argument("--pretrained_weights", type=str, help="if specified starts from checkpoint model")
    parser.add_argument("--n_cpu", type=int, default=8, help="number of cpu threads to use during batch generation")
    parser.add_argument("--img_size", type=int, default=416, help="size of each image dimension")
    parser.add_argument("--checkpoint_interval", type=int, default=1, help="interval between saving model weights")
    parser.add_argument("--evaluation_interval", type=int, default=1, help="interval evaluations on validation set")
    parser.add_argument("--compute_map", default=False, help="if True computes mAP every tenth batch")
    parser.add_argument("--multiscale_training", default=True, help="allow for multi-scale training")
    parser.add_argument("--verbose", "-v", default=False, action="store_true", help="Makes the training more verbose")
    parser.add_argument(
        "--logdir", type=str, default="logs", help="Defines the directory where the training log files are stored"
    )
    parser.add_argument("--conf_thres", type=float, default=0.8, help="object confidence threshold")
    parser.add_argument("--nms_thres", type=float, default=0.4, help="iou thresshold for non-maximum suppression")
    parser.add_argument("--device", type=str, default="auto", help="specify device")
    parser.add_argument(
        "--max_files", type=int, default=None, help="specify dataset size (used to check if model can overfit)"
    )
    parser.add_argument("--weights_path", type=str, default=checkpoint_path, help="path to weights file")
    parser.add_argument("--datset_split", type=float, default=0.95, help="how much data should be trained on")
    parser.add_argument("--fast", default=False, action="store_true", help="skip checkpoints and acc while training")
    parser.add_argument("--name", type=str, default="", help="prefix of checkpoint filename")
    parser.add_argument("--weighted", default=False, action="store_true", help="train with weighted dataset")

    return parser


def get_opts():
    parser = get_opts_raw()
    opt = parser.parse_args()
    return opt


def get_model(opt, device):
    model = Darknet(opt.model_def).to(device)
    model.apply(weights_init_normal)

    if os.path.isfile(opt.weights_path):
        if opt.weights_path.endswith(".weights"):
            # Load darknet weights
            model.load_darknet_weights(opt.weights_path)
        else:
            # Load checkpoint weights
            model.load_state_dict(torch.load(opt.weights_path))

    if opt.pretrained_weights:
        if opt.pretrained_weights.endswith(".pth"):
            model.load_state_dict(torch.load(opt.pretrained_weights))
        else:
            model.load_darknet_weights(opt.pretrained_weights)
    return model


if __name__ == "__main__":
    opt = get_opts()
    logger = Logger(opt.logdir)

    assert opt.device in ["cuda", "cpu", "auto"], "invalid device"
    device = (
        torch.device("cuda" if torch.cuda.is_available() else "cpu")
        if opt.device == "auto"
        else torch.device(opt.device)
    )

    os.makedirs("output", exist_ok=True)
    os.makedirs("checkpoints", exist_ok=True)

    # Get data configuration
    data_config = parse_data_config(opt.data_config)
    current_dir = os.path.dirname(os.path.abspath(__file__))
    root_path = os.path.join(current_dir, data_config["root_folder"])
    image_path = os.path.join(root_path, "images")
    class_names = os.path.join(root_path, "classes.names")
    filters = load_filters(data_config.get("filter", None), current_dir)

    wandb.init(project="yolov3_" + os.path.dirname(root_path).split("/")[-1], entity="sigurd-og-brage-bachelor")

    assert os.path.isdir(image_path), "{path} is not a directory".format(path=image_path)
    assert os.path.isfile(class_names), "{path} is not a file".format(path=class_names)
    class_names = load_classes(class_names)

    print("training on ", class_names)

    model = get_model(opt, device)

    max_files = opt.max_files
    dataset_train, dataset_validator = get_train_and_validate_dataset(
        image_path=image_path,
        split=opt.datset_split,
        opt=opt,
        transform=DEFAULT_TRANSFORMS,
        max_files=max_files,
        filters=filters,
    )

    dataloader = torch.utils.data.DataLoader(
        dataset_train,
        batch_size=opt.batch_size,
        shuffle=True,
        num_workers=opt.n_cpu,
        pin_memory=True,
        collate_fn=dataset_train.collate_fn,
    )
    validator_dataloader = torch.utils.data.DataLoader(
        dataset_validator,
        batch_size=opt.batch_size,
        shuffle=True,
        num_workers=opt.n_cpu,
        pin_memory=True,
        collate_fn=dataset_train.collate_fn,
    )

    optimizer = torch.optim.Adam(model.parameters())

    metrics = [
        "grid_size",
        "loss",
        "x",
        "y",
        "w",
        "h",
        "conf",
        "cls",
        "cls_acc",
        "recall50",
        "recall75",
        "precision",
        "conf_obj",
        "conf_noobj",
    ]
    accuracy = []
    prefix_name = opt.name
    prefix_name += "_" if len(prefix_name) > 0 else ""
    for epoch in range(opt.epochs):
        model.train()
        accumulated_loss = 0
        start_time = time.time()
        for batch_i, (file_img, imgs, targets) in enumerate(tqdm.tqdm(dataloader, desc=f"Training Epoch {epoch}")):
            batches_done = len(dataloader) * epoch + batch_i
            imgs = Variable(imgs.to(device))
            targets = Variable(targets.to(device), requires_grad=False)

            loss, outputs = model(imgs, targets)
            accumulated_loss += loss.item()
            loss.backward()

            if batches_done % opt.gradient_accumulations == 0:
                optimizer.step()
                optimizer.zero_grad()
                # wandb.log({"loss": accumulated_loss})
                accumulated_loss = 0
            torch.cuda.empty_cache()

        if not opt.fast or epoch % 100 == 0:
            torch.save(
                model.state_dict(),
                os.path.join(current_dir, "checkpoints/{name}latest_{epoch}.pth".format(name=prefix_name, epoch=epoch)),
            )
            checkpoint = os.path.join(current_dir, "checkpoints/{name}latest_checkpoint.pth".format(name=prefix_name))
            torch.save(model.state_dict(), checkpoint)
            print("stored checkpoint ", checkpoint)

        if not opt.fast or (epoch + 1) == opt.epochs:
            accuracy_out, (precision, recall, AP, f1, ap_class) = get_accracy_of_model(
                model, dataset_validator, validator_dataloader, opt, class_names, epoch=epoch
            )
            if ap_class is None:
                continue

            for i, c in enumerate(ap_class):
                wandb.log({"precision " + str(class_names[c]): precision[i]})
                wandb.log({"f1 " + str(class_names[c]): f1[i]})
                wandb.log({"recall " + str(class_names[c]): recall[i]})
                wandb.log({"average precision " + str(class_names[c]): AP[i]})
                wandb.log({"average precision per class " + str(class_names[c]): ap_class[i]})

            accuracy.append(accuracy_out)

            wandb.log({"acc": accuracy[-1]})

            wandb.run.summary["best_accuracy"] = max(accuracy)
            wandb.run.summary["epochs"] = epoch
            wandb.run.summary["fields"] = len(class_names)
            wandb.run.summary["dataset_size"] = len(dataloader)
            wandb.run.summary["creditor-ids"] = "-".join(dataset_train.get_creditor_ids())
